from __future__ import annotations

__all__ = ["BasePreprocessor", "is_preprocessor_config", "setup_preprocessor"]

import logging
from abc import ABC

from objectory import AbstractFactory
from objectory.utils import is_object_config
from pandas import DataFrame

logger = logging.getLogger(__name__)


class BasePreprocessor(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to preprocess a DataFrame.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.preprocessor import ToNumericPreprocessor
        >>> preprocessor = ToNumericPreprocessor(columns=["col1", "col3"])
        >>> preprocessor
        ToNumericPreprocessor(columns=('col1', 'col3'))
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> df.dtypes
        col1     int64
        col2    object
        col3    object
        col4    object
        dtype: object
        >>> df = preprocessor.preprocess(df)
        >>> df.dtypes
        col1     int64
        col2    object
        col3     int64
        col4    object
        dtype: object
    """

    def preprocess(self, df: DataFrame) -> DataFrame:
        r"""Preprocesses the data in the DataFrame.

        Args:
        ----
            df (``pandas.DataFrame``): Specifies the DataFrame
                to preprocess.

        Returns:
        -------
            ``pandas.DataFrame``: The preprocessed DataFrame.

        Example usage:

        .. code-block:: pycon

            >>> import pandas as pd
            >>> from flamme.preprocessor import ToNumericPreprocessor
            >>> preprocessor = ToNumericPreprocessor(columns=["col1", "col3"])
            >>> df = pd.DataFrame(
            ...     {
            ...         "col1": [1, 2, 3, 4, 5],
            ...         "col2": ["1", "2", "3", "4", "5"],
            ...         "col3": ["1", "2", "3", "4", "5"],
            ...         "col4": ["a", "b", "c", "d", "e"],
            ...     }
            ... )
            >>> df.dtypes
            col1     int64
            col2    object
            col3    object
            col4    object
            dtype: object
            >>> df = preprocessor.preprocess(df)
            >>> df.dtypes
            col1     int64
            col2    object
            col3     int64
            col4    object
            dtype: object
        """


def is_preprocessor_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BasePreprocessor``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BasePreprocessor`` object.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.preprocessor import is_preprocessor_config
        >>> is_preprocessor_config(
        ...     {
        ...         "_target_": "flamme.preprocessor.ToNumericPreprocessor",
        ...         "columns": ["col1", "col3"],
        ...     }
        ... )
        True
    """
    return is_object_config(config, BasePreprocessor)


def setup_preprocessor(
    preprocessor: BasePreprocessor | dict,
) -> BasePreprocessor:
    r"""Sets up an preprocessor.

    The preprocessor is instantiated from its configuration
    by using the ``BasePreprocessor`` factory function.

    Args:
    ----
        preprocessor (``BasePreprocessor`` or dict): Specifies an
            preprocessor or its configuration.

    Returns:
    -------
        ``BasePreprocessor``: An instantiated preprocessor.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.preprocessor import setup_preprocessor
        >>> preprocessor = setup_preprocessor(
        ...     {
        ...         "_target_": "flamme.preprocessor.ToNumericPreprocessor",
        ...         "columns": ["col1", "col3"],
        ...     }
        ... )
        >>> preprocessor
        ToNumericPreprocessor(columns=('col1', 'col3'))
    """
    if isinstance(preprocessor, dict):
        logger.info("Initializing an preprocessor from its configuration... ")
        preprocessor = BasePreprocessor.factory(**preprocessor)
    if not isinstance(preprocessor, BasePreprocessor):
        logger.warning(f"preprocessor is not a `BasePreprocessor` (received: {type(preprocessor)})")
    return preprocessor



from __future__ import annotations

__all__ = ["save_text"]

import logging
from pathlib import Path

logger = logging.getLogger(__name__)


def save_text(to_save: str, path: Path) -> None:
    r"""Saves the given data in a text file.

    Args:
    ----
        to_save: Specifies the data to write in a text file.
        path (``pathlib.Path``): Specifies the path where to write the
            text file.

    Example usage:

    .. code-block:: pycon

        >>> from pathlib import Path
        >>> from flamme.utils.io import save_text
        >>> save_text("abc", Path("/path/to/data.txt"))  # xdoctest: +SKIP()
    """
    logger.debug(f"write data in a text file: {path}")
    path.parents[0].mkdir(exist_ok=True, parents=True)
    # Save to tmp, then commit by moving the file in case the job gets
    # interrupted while writing the file
    tmp_path = path.parents[0].joinpath(f"{path.name}.tmp")
    with Path.open(tmp_path, mode="w") as file:
        file.write(to_save)
    tmp_path.rename(path)



from __future__ import annotations

__all__ = ["ColumnDiscreteAnalyzer", "ColumnTemporalDiscreteAnalyzer"]

import logging
from collections import Counter

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import (
    ColumnDiscreteSection,
    ColumnTemporalDiscreteSection,
    EmptySection,
)

logger = logging.getLogger(__name__)


class ColumnDiscreteAnalyzer(BaseAnalyzer):
    r"""Implements a discrete distribution analyzer.

    Args:
    ----
        column (str): Specifies the column to analyze.
        dropna (bool, optional): If ``True``, the NaN values are not
            included in the analysis. Default: ``False``
        max_rows (int, optional): Specifies the maximum number of rows
            to show in the table. Default: ``20``

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import ColumnDiscreteAnalyzer
        >>> analyzer = ColumnDiscreteAnalyzer(column="str")
        >>> analyzer
        ColumnDiscreteAnalyzer(column=str, dropna=False, max_rows=20)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(self, column: str, dropna: bool = False, max_rows: int = 20) -> None:
        self._column = column
        self._dropna = bool(dropna)
        self._max_rows = max_rows

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"dropna={self._dropna}, max_rows={self._max_rows})"
        )

    def analyze(self, df: DataFrame) -> ColumnDiscreteSection | EmptySection:
        logger.info(f"Analyzing the discrete distribution of {self._column}")
        if self._column not in df:
            logger.info(
                f"Skipping discrete distribution analysis of column {self._column} "
                f"because it is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return ColumnDiscreteSection(
            counter=Counter(df[self._column].value_counts(dropna=self._dropna).to_dict()),
            null_values=df[self._column].isnull().sum(),
            column=self._column,
            max_rows=self._max_rows,
        )


class ColumnTemporalDiscreteAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of
    discrete values.

    Args:
    ----
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import ColumnTemporalDiscreteAnalyzer
        >>> analyzer = ColumnTemporalDiscreteAnalyzer(
        ...     column="str", dt_column="datetime", period="M"
        ... )
        >>> analyzer
        ColumnTemporalDiscreteAnalyzer(column=str, dt_column=datetime, period=M)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...         "datetime": pd.to_datetime(
        ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
        ...         ),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(self, column: str, dt_column: str, period: str) -> None:
        self._column = column
        self._dt_column = dt_column
        self._period = period

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"dt_column={self._dt_column}, period={self._period})"
        )

    def analyze(self, df: DataFrame) -> ColumnTemporalDiscreteSection | EmptySection:
        logger.info(
            f"Analyzing the temporal discrete distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        if self._column not in df:
            logger.info(
                "Skipping temporal discrete distribution analysis because the column "
                f"({self._column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        if self._dt_column not in df:
            logger.info(
                "Skipping temporal discrete distribution analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return ColumnTemporalDiscreteSection(
            column=self._column,
            df=df,
            dt_column=self._dt_column,
            period=self._period,
        )



from __future__ import annotations

__all__ = ["GlobalTemporalNullValueSection"]

import logging
from collections.abc import Sequence

import numpy as np
import plotly
import plotly.graph_objects as go
from jinja2 import Template
from pandas import DataFrame
from plotly.subplots import make_subplots

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)

logger = logging.getLogger(__name__)


class GlobalTemporalNullValueSection(BaseSection):
    r"""Implements a section to analyze the temporal distribution of null
    values for all columns.

    Args:
    ----
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        figsize (``tuple`` or list , optional): Specifies the figure
            size in pixels. The first dimension is the width and the
            second is the height. Default: ``(None, None)``
    """

    def __init__(
        self,
        df: DataFrame,
        dt_column: str,
        period: str,
        figsize: tuple[int | None, int | None] | list[int | None] = (None, None),
    ) -> None:
        if dt_column not in df:
            raise ValueError(
                f"Datetime column {dt_column} is not in the DataFrame "
                f"(columns:{sorted(df.columns)})"
            )

        self._df = df
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    @property
    def df(self) -> DataFrame:
        r"""``pandas.DataFrame``: The DataFrame to analyze."""
        return self._df

    @property
    def dt_column(self) -> str:
        r"""str: The datetime column."""
        return self._dt_column

    @property
    def period(self) -> str:
        r"""str: The temporal period used to analyze the data."""
        return self._period

    @property
    def figsize(self) -> tuple[int, int]:
        r"""tuple: The individual figure size in pixels. The first
        dimension is the width and the second is the height."""
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            "Rendering the temporal distribution of null values for all columns "
            f"| datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "dt_column": self._dt_column,
                "figure": create_temporal_null_figure(
                    df=self._df,
                    dt_column=self._dt_column,
                    period=self._period,
                    figsize=self._figsize,
                ),
                "table": create_temporal_null_table(
                    df=self._df,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of null values in all columns.
The column <em>{{dt_column}}</em> is used as temporal column.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_null_figure(
    df: DataFrame,
    dt_column: str,
    period: str,
    figsize: tuple[int | None, int | None] | list[int | None] = (None, None),
) -> str:
    r"""Creates a HTML representation of a figure with the temporal null
    value distribution.

    Args:
    ----
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        figsize (``tuple`` or list , optional): Specifies the figure
            size in pixels. The first dimension is the width and the
            second is the height. Default: ``(None, None)``

    Returns:
    -------
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return ""

    num_nulls, total, labels = prepare_data(df=df, dt_column=dt_column, period=period)

    fig = make_subplots(rows=1, cols=1, specs=[[{"secondary_y": True}]])
    fig.add_trace(
        go.Bar(
            x=labels,
            y=total,
            marker=dict(color="rgba(0, 191, 255, 0.9)"),
        ),
        secondary_y=False,
    )
    fig.add_trace(
        go.Bar(
            x=labels,
            y=num_nulls,
            marker=dict(color="rgba(255, 191, 0, 0.9)"),
        ),
        secondary_y=False,
    )
    fig.add_trace(
        go.Scatter(
            x=labels,
            y=num_nulls / total,
            marker=dict(color="rgba(0, 71, 171, 0.9)"),
        ),
        secondary_y=True,
    )
    fig.update_yaxes(
        title_text=(
            '<span style="color:RGB(255, 191, 0)">null</span>/'
            '<span style="color:RGB(0, 191, 255)">total</span>'
        ),
        secondary_y=False,
    )
    fig.update_yaxes(
        title_text='<span style="color:RGB(0, 71, 171)">percentage</span>',
        secondary_y=True,
    )
    fig.update_layout(
        height=figsize[1],
        width=figsize[0],
        showlegend=False,
        barmode="overlay",
    )
    return plotly.io.to_html(fig, full_html=False)


def create_temporal_null_table(df: DataFrame, dt_column: str, period: str) -> str:
    r"""Creates a HTML representation of a table with the temporal
    distribution of null values.

    Args:
    ----
        df (``DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Returns:
    -------
        str: The HTML representation of the table.
    """
    if df.shape[0] == 0:
        return ""
    num_nulls, totals, labels = prepare_data(df=df, dt_column=dt_column, period=period)
    rows = []
    for label, num_null, total in zip(labels, num_nulls, totals):
        rows.append(create_temporal_null_table_row(label=label, num_nulls=num_null, total=total))
    return Template(
        """
<details>
    <summary>Statistics per period</summary>

    <p>The following table shows some statistics for each period.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>period</th>
                <th>number of null values</th>
                <th>number of non-null values</th>
                <th>total number of values</th>
                <th>percentage of null values</th>
                <th>percentage of non-null values</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "period": period})


def create_temporal_null_table_row(label: str, num_nulls: int, total: int) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
    ----
        row ("pd.core.frame.Pandas"): Specifies a DataFrame row.

    Returns:
    -------
        str: The HTML code of a row.
    """
    num_non_nulls = total - num_nulls
    return Template(
        """<tr>
    <th>{{label}}</th>
    <td {{num_style}}>{{num_nulls}}</td>
    <td {{num_style}}>{{num_non_nulls}}</td>
    <td {{num_style}}>{{total}}</td>
    <td {{num_style}}>{{num_nulls_pct}}</td>
    <td {{num_style}}>{{num_non_nulls_pct}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "label": label,
            "num_nulls": f"{num_nulls:,}",
            "num_non_nulls": f"{num_non_nulls:,}",
            "total": f"{total:,}",
            "num_nulls_pct": f"{100 * num_nulls / total:.2f}%",
            "num_non_nulls_pct": f"{100 * num_non_nulls / total:.2f}%",
        }
    )


def prepare_data(
    df: DataFrame,
    dt_column: str,
    period: str,
) -> tuple[np.ndarray, np.ndarray, list]:
    r"""Prepares the data to create the figure and table.

    Args:
    ----
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Returns:
    -------
        tuple: A tuple with 3 values. The first value is a numpy NDArray
            that contains the number of null values per period. The
            second value is a numpy NDArray that contains the total
            number of values. The third value is a list that contains
            the label of each period.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.section.null_temp_global import prepare_data
        >>> num_nulls, total, labels = prepare_data(
        ...     df=pd.DataFrame(
        ...         {
        ...             "col1": np.array([np.nan, 1, 0, 1]),
        ...             "col2": np.array([np.nan, 1, 0, np.nan]),
        ...             "datetime": pd.to_datetime(
        ...                 ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
        ...             ),
        ...         }
        ...     ),
        ...     dt_column="datetime",
        ...     period="M",
        ... )
        >>> num_nulls
        array([2, 0, 0, 1])
        >>> total
        array([2, 2, 2, 2])
        >>> labels
        ['2020-01', '2020-02', '2020-03', '2020-04']
    """
    df = df.copy()
    columns = df.columns.tolist()
    columns.remove(dt_column)
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period)
    df.loc[:, columns] = df[columns].isnull()

    df_num_nulls = df.groupby(dt_col)[columns].sum().sum(axis=1).sort_index()
    df_total = df.groupby(dt_col)[columns].count().sum(axis=1).sort_index()
    labels = [str(dt) for dt in df_num_nulls.index]
    return df_num_nulls.to_numpy().astype(int), df_total.to_numpy().astype(int), labels



from __future__ import annotations

__all__ = [
    "BasePreprocessor",
    "ColumnSelectionPreprocessor",
    "NullColumnPreprocessor",
    "SequentialPreprocessor",
    "StripStrPreprocessor",
    "ToDatetimePreprocessor",
    "ToNumericPreprocessor",
    "is_preprocessor_config",
    "setup_preprocessor",
]

from flamme.preprocessor.base import (
    BasePreprocessor,
    is_preprocessor_config,
    setup_preprocessor,
)
from flamme.preprocessor.column import ColumnSelectionPreprocessor
from flamme.preprocessor.datetime import ToDatetimePreprocessor
from flamme.preprocessor.null import NullColumnPreprocessor
from flamme.preprocessor.numeric import ToNumericPreprocessor
from flamme.preprocessor.sequential import SequentialPreprocessor
from flamme.preprocessor.str import StripStrPreprocessor



from __future__ import annotations

__all__ = ["ColumnContinuousSection"]

import logging
from collections.abc import Sequence

import plotly
import plotly.express as px
from jinja2 import Template
from pandas import Series

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.range import find_range

logger = logging.getLogger(__name__)


class ColumnContinuousSection(BaseSection):
    r"""Implements a section that analyzes a continuous distribution of
    values.

    Args:
    ----
        series (``pandas.Series``): Specifies the series/column to
            analyze.
        column (str): Specifies the column name.
        nbins (int or None, optional): Specifies the number of bins in
            the histogram. Default: ``None``
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``
        xmin (float or str or None, optional): Specifies the minimum
            value of the range or its associated quantile.
            ``q0.1`` means the 10% quantile. ``0`` is the minimum
            value and ``1`` is the maximum value. Default: ``None``
        xmax (float or str or None, optional): Specifies the maximum
            value of the range or its associated quantile.
            ``q0.9`` means the 90% quantile. ``0`` is the minimum
            value and ``1`` is the maximum value. Default: ``None``
    """

    def __init__(
        self,
        series: Series,
        column: str,
        nbins: int | None = None,
        log_y: bool = False,
        xmin: float | str | None = None,
        xmax: float | str | None = None,
    ) -> None:
        self._series = series
        self._column = column
        self._nbins = nbins
        self._log_y = log_y
        self._xmin = xmin
        self._xmax = xmax

    @property
    def column(self) -> str:
        return self._column

    @property
    def log_y(self) -> bool:
        return self._log_y

    @property
    def nbins(self) -> int | None:
        return self._nbins

    @property
    def series(self) -> Series:
        return self._series

    @property
    def xmin(self) -> float | str | None:
        return self._xmin

    @property
    def xmax(self) -> float | str | None:
        return self._xmax

    def get_statistics(self) -> dict:
        stats = {
            "count": int(self._series.shape[0]),
            "num_nulls": int(self._series.isnull().sum()),
            "nunique": self._series.nunique(dropna=False),
        }
        stats["num_non_nulls"] = stats["count"] - stats["num_nulls"]
        if stats["num_non_nulls"] > 0:
            stats |= (
                self._series.dropna()
                .agg(
                    {
                        "mean": "mean",
                        "median": "median",
                        "min": "min",
                        "max": "max",
                        "std": "std",
                        "q01": lambda x: x.quantile(0.01),
                        "q05": lambda x: x.quantile(0.05),
                        "q10": lambda x: x.quantile(0.1),
                        "q25": lambda x: x.quantile(0.25),
                        "q75": lambda x: x.quantile(0.75),
                        "q90": lambda x: x.quantile(0.9),
                        "q95": lambda x: x.quantile(0.95),
                        "q99": lambda x: x.quantile(0.99),
                    }
                )
                .to_dict()
            )
        else:
            stats |= {
                "mean": float("nan"),
                "median": float("nan"),
                "min": float("nan"),
                "max": float("nan"),
                "std": float("nan"),
                "q01": float("nan"),
                "q05": float("nan"),
                "q10": float("nan"),
                "q25": float("nan"),
                "q75": float("nan"),
                "q90": float("nan"),
                "q95": float("nan"),
                "q99": float("nan"),
            }
        return stats

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the continuous distribution of {self._column}")
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * stats['num_nulls'] / stats['count']:.2f}" if stats["count"] > 0 else "N/A"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "table": create_stats_table(stats=stats, column=self._column),
                "total_values": f"{stats['count']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{stats['num_nulls']:,}",
                "null_values_pct": null_values_pct,
                "figure": create_histogram_figure(
                    series=self._series,
                    column=self._column,
                    nbins=self._nbins,
                    log_y=self._log_y,
                    xmin=self._xmin,
                    xmax=self._xmax,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the discrete distribution of values for column {{column}}.

<ul>
  <li> total values: {{total_values}} </li>
  <li> number of unique values: {{unique_values}} </li>
  <li> number of null values: {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
</ul>

{{figure}}
{{table}}
<p style="margin-top: 1rem;">
"""


def create_histogram_figure(
    series: Series,
    column: str,
    nbins: int | None = None,
    log_y: bool = False,
    xmin: float | str | None = None,
    xmax: float | str | None = None,
) -> str:
    r"""Creates the HTML code of a figure.

    Args:
    ----
        row (``pandas.Series``): Specifies the series of data.
        column (str): Specifies the column name.
        nbins (int or None, optional): Specifies the number of bins in
            the histogram. Default: ``None``
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``

    Returns:
    -------
        str: The HTML code of the figure.
    """
    array = series.to_numpy()
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    fig = px.histogram(
        array,
        marginal="box",
        nbins=nbins,
        title=f"Distribution of values for column {column}",
        labels={"x": "value", "y": "count"},
        log_y=log_y,
        range_x=[xmin, xmax],
    )
    fig.update_layout(showlegend=False)

    return plotly.io.to_html(fig, full_html=False)


def create_stats_table(stats: dict, column: str) -> str:
    r"""Creates the HTML code of the table with statistics.

    Args:
    ----
        stats (dict): Specifies a dictionary with the statistics.
        column (str): Specifies the column name.

    Returns:
    -------
        str: The HTML code of the table.
    """
    return Template(
        """
<details>
    <summary>Statistics</summary>

    <p>The following table shows some statistics about the distribution for column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr><th>stat</th><th>value</th></tr>
        </thead>
        <tbody class="tbody table-group-divider">
            <tr><th>count</th><td {{num_style}}>{{count}}</td></tr>
            <tr><th>mean</th><td {{num_style}}>{{mean}}</td></tr>
            <tr><th>std</th><td {{num_style}}>{{std}}</td></tr>
            <tr><th>min</th><td {{num_style}}>{{min}}</td></tr>
            <tr><th>quantile 1%</th><td {{num_style}}>{{q01}}</td></tr>
            <tr><th>quantile 5%</th><td {{num_style}}>{{q05}}</td></tr>
            <tr><th>quantile 10%</th><td {{num_style}}>{{q10}}</td></tr>
            <tr><th>quantile 25%</th><td {{num_style}}>{{q25}}</td></tr>
            <tr><th>median</th><td {{num_style}}>{{median}}</td></tr>
            <tr><th>quantile 75%</th><td {{num_style}}>{{q75}}</td></tr>
            <tr><th>quantile 90%</th><td {{num_style}}>{{q90}}</td></tr>
            <tr><th>quantile 95%</th><td {{num_style}}>{{q95}}</td></tr>
            <tr><th>quantile 99%</th><td {{num_style}}>{{q99}}</td></tr>
            <tr><th>max</th><td {{num_style}}>{{max}}</td></tr>
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render(
        {
            "column": column,
            "num_style": 'style="text-align: right;"',
            "count": f"{stats['count']:,}",
            "mean": f"{stats['mean']:,.4f}",
            "median": f"{stats['median']:,.4f}",
            "min": f"{stats['min']:,.4f}",
            "max": f"{stats['max']:,.4f}",
            "std": f"{stats['std']:,.4f}",
            "q01": f"{stats['q01']:,.4f}",
            "q05": f"{stats['q05']:,.4f}",
            "q10": f"{stats['q10']:,.4f}",
            "q25": f"{stats['q25']:,.4f}",
            "q75": f"{stats['q75']:,.4f}",
            "q90": f"{stats['q90']:,.4f}",
            "q95": f"{stats['q95']:,.4f}",
            "q99": f"{stats['q99']:,.4f}",
        }
    )



from __future__ import annotations

__all__ = ["ColumnTemporalContinuousSection"]

import logging
from collections.abc import Sequence

import pandas as pd
import plotly
import plotly.express as px
from jinja2 import Template
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)

logger = logging.getLogger(__name__)


class ColumnTemporalContinuousSection(BaseSection):
    r"""Implements a section that analyzes the temporal distribution of a
    column with continuous values.

    Args:
    ----
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column of the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``
    """

    def __init__(
        self,
        df: DataFrame,
        column: str,
        dt_column: str,
        period: str,
        log_y: bool = False,
    ) -> None:
        self._df = df
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._log_y = log_y

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def log_y(self) -> bool:
        return self._log_y

    @property
    def period(self) -> str:
        return self._period

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Rendering the temporal continuous distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "period": self._period,
                "figure": create_temporal_figure(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                    log_y=self._log_y,
                ),
                "table": create_temporal_table(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of column {{column}} by using the column {{dt_column}}.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_figure(
    df: DataFrame, column: str, dt_column: str, period: str, log_y: bool = False
) -> str:
    r"""Creates a HTML representation of a figure with the temporal value
    distribution.

    Args:
    ----
        df (``DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``

    Returns:
    -------
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return ""
    df = df[[column, dt_column]].copy()
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period).astype(str)

    fig = px.box(
        df,
        x=dt_col,
        y=column,
        title=f"Distribution of values for column {column}",
        labels={column: "value", dt_col: "time"},
        points="outliers",
        log_y=log_y,
    )
    return plotly.io.to_html(fig, full_html=False)


def create_temporal_table(df: DataFrame, column: str, dt_column: str, period: str) -> str:
    r"""Creates a HTML representation of a table with some statistics
    about the temporal value distribution.

    Args:
    ----
        df (``DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Returns:
    -------
        str: The HTML representation of the table.
    """
    if df.shape[0] == 0:
        return ""
    df = df[[column, dt_column]].copy()
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period)
    df_stats = (
        df.groupby(dt_col)[column]
        .agg(
            [
                "count",
                "mean",
                "median",
                "min",
                "max",
                "std",
                ("q01", lambda x: x.quantile(0.01)),
                ("q10", lambda x: x.quantile(0.1)),
                ("q25", lambda x: x.quantile(0.25)),
                ("q75", lambda x: x.quantile(0.75)),
                ("q90", lambda x: x.quantile(0.9)),
                ("q99", lambda x: x.quantile(0.99)),
            ]
        )
        .sort_index()
    )

    rows = []
    for row in df_stats.itertuples():
        rows.append(create_temporal_table_row(row))
    return Template(
        """
<details>
    <summary>Statistics per period</summary>

    <p>The following table shows some statistics for each period of column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>column</th>
                <th>count</th>
                <th>mean</th>
                <th>std</th>
                <th>min</th>
                <th>quantile 1%</th>
                <th>quantile 10%</th>
                <th>quantile 25%</th>
                <th>median</th>
                <th>quantile 75%</th>
                <th>quantile 90%</th>
                <th>quantile 99%</th>
                <th>max</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "column": column, "period": period})


def create_temporal_table_row(row: pd.core.frame.Pandas) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
    ----
        row ("pd.core.frame.Pandas"): Specifies a DataFrame row.

    Returns:
    -------
        str: The HTML code of a row.
    """
    return Template(
        """<tr>
    <th>{{datetime}}</th>
    <td {{num_style}}>{{count}}</td>
    <td {{num_style}}>{{mean}}</td>
    <td {{num_style}}>{{std}}</td>
    <td {{num_style}}>{{min}}</td>
    <td {{num_style}}>{{q01}}</td>
    <td {{num_style}}>{{q10}}</td>
    <td {{num_style}}>{{q25}}</td>
    <td {{num_style}}>{{median}}</td>
    <td {{num_style}}>{{q75}}</td>
    <td {{num_style}}>{{q90}}</td>
    <td {{num_style}}>{{q99}}</td>
    <td {{num_style}}>{{max}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "datetime": row.Index,
            "count": f"{row.count:,}",
            "mean": f"{row.mean:,.4f}",
            "median": f"{row.median:,.4f}",
            "min": f"{row.min:,.4f}",
            "max": f"{row.max:,.4f}",
            "std": f"{row.std:,.4f}",
            "q01": f"{row.q01:,.4f}",
            "q10": f"{row.q10:,.4f}",
            "q25": f"{row.q25:,.4f}",
            "q75": f"{row.q75:,.4f}",
            "q90": f"{row.q90:,.4f}",
            "q99": f"{row.q99:,.4f}",
        }
    )



from __future__ import annotations

__all__ = ["valid_h_tag", "tags2id", "tags2title", "render_html_toc"]

from collections.abc import Sequence

GO_TO_TOP = '<a href="#">Go to top</a>'


def tags2id(tags: Sequence[str]) -> str:
    r"""Converts a sequence of tags to a string that can be used as ID in
    a HTML file.

    Args:
    ----
        tags (``Sequence``): Specifies the sequence of tags.

    Returns:
    -------
        str: The generated ID from the tags.
    """
    return "-".join(tags).replace(" ", "-").lower()


def tags2title(tags: Sequence[str]) -> str:
    r"""Converts a sequence of tags to a string that can be used as
    title.

    Args:
    ----
        tags (``Sequence``): Specifies the sequence of tags.

    Returns:
    -------
        str: The generated title from the tags.
    """
    return " | ".join(tags[::-1])


def valid_h_tag(index: int) -> int:
    r"""Computes a valid number of a h HTML tag.

    Args:
    ----
        index (int): Specifies the original value.

    Returns:
    -------
        int: A valid value.
    """
    return max(1, min(6, index))


def render_html_toc(
    number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
) -> str:
    r"""Renders the HTML table of content (TOC) associated to the
    section.

    Args:
    ----
        number (str, optional): Specifies the section number
            associated to the section. Default: ""
        tags (``Sequence``, optional): Specifies the tags
            associated to the section. Default: ``()``
        depth (int, optional): Specifies the depth in the report.
            Default: ``0``

    Returns:
    -------
        str: The HTML table of content associated to the section.
    """
    if depth >= max_depth:
        return ""
    tag = tags[-1] if tags else ""
    return f'<li><a href="#{tags2id(tags)}">{number} {tag}</a></li>'



from __future__ import annotations

__all__ = ["Reporter"]

import logging
from pathlib import Path

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer
from flamme.ingestor.base import BaseIngestor, setup_ingestor
from flamme.preprocessor.base import BasePreprocessor, setup_preprocessor
from flamme.reporter.base import BaseReporter
from flamme.reporter.utils import create_html_report
from flamme.utils.io import save_text
from flamme.utils.path import sanitize_path

logger = logging.getLogger(__name__)


class Reporter(BaseReporter):
    r"""Implements a simple reporter.

    Args:
    ----
        ingestor (``BaseIngestor`` or dict): Specifies the ingestor
            or its configuration.
        preprocessor (``BasePreprocessor`` or dict): Specifies the
            DataFrame preprocessor or its configuration.
        ingestor (``BaseAnalyzer`` or dict): Specifies the analyzer
            or its configuration.
        report_path (``Path`` or str): Specifies the path where to
            save the HTML report.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.analyzer import NullValueAnalyzer
        >>> from flamme.ingestor import ParquetIngestor
        >>> from flamme.preprocessor import SequentialPreprocessor
        >>> from flamme.reporter import Reporter
        >>> reporter = Reporter(
        ...     ingestor=ParquetIngestor("/path/to/data.parquet"),
        ...     preprocessor=SequentialPreprocessor(preprocessors=[]),
        ...     analyzer=NullValueAnalyzer(),
        ...     report_path="/path/to/report.html",
        ... )
        >>> report = reporter.compute()  # doctest: +SKIP
    """

    def __init__(
        self,
        ingestor: BaseIngestor | dict,
        preprocessor: BasePreprocessor | dict,
        analyzer: BaseAnalyzer | dict,
        report_path: Path | str,
    ) -> None:
        self._ingestor = setup_ingestor(ingestor)
        logger.info(f"ingestor:\n{ingestor}")
        self._preprocessor = setup_preprocessor(preprocessor)
        logger.info(f"preprocessor:\n{preprocessor}")
        self._analyzer = setup_analyzer(analyzer)
        logger.info(f"analyzer:\n{analyzer}")
        self._report_path = sanitize_path(report_path)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "ingestor": self._ingestor,
                    "preprocessor": self._preprocessor,
                    "analyzer": self._analyzer,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute(self) -> None:
        logger.info("Ingesting the DataFrame...")
        df = self._ingestor.ingest()
        logger.info(f"Preprocessing the DataFrame {df.shape}...")
        df = self._preprocessor.preprocess(df)
        logger.info(f"Analyzing the DataFrame {df.shape}...")
        section = self._analyzer.analyze(df)
        logger.info("Creating the HTML report...")
        report = create_html_report(
            toc=section.render_html_toc(max_depth=6),
            body=section.render_html_body(),
        )
        logger.info(f"Saving HTML report at {self._report_path}...")
        save_text(report, self._report_path)



from __future__ import annotations

__all__ = ["NullColumnPreprocessor"]


from pandas import DataFrame

from flamme.preprocessor.base import BasePreprocessor
from flamme.utils.null import compute_null_per_col


class NullColumnPreprocessor(BasePreprocessor):
    r"""Implements a preprocessor to remove the columns that have too
    many null values.

    Args:
    ----
        threshold (float): Specifies the maximum percentage of null
            values to keep columns. If the proportion of null vallues
            is greater or equal to this threshold value, the column
            is removed. If set to ``1.0``, it removes all the columns
            that have only null values.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.preprocessor import ToDatetimePreprocessor
        >>> preprocessor = NullColumnPreprocessor(threshold=1.0)
        >>> preprocessor
        NullColumnPreprocessor(threshold=1.0)
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": ["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", None],
        ...         "col2": [1, None, 3, None, 5],
        ...         "col3": [None, None, None, None, None],
        ...     }
        ... )
        >>> df = preprocessor.preprocess(df)
        >>> df
                 col1  col2
        0    2020-1-1   1.0
        1    2020-1-2   NaN
        2   2020-1-31   3.0
        3  2020-12-31   NaN
        4        None   5.0
    """

    def __init__(self, threshold: float) -> None:
        self._threshold = float(threshold)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(threshold={self._threshold})"

    def preprocess(self, df: DataFrame) -> DataFrame:
        if df.shape[0] == 0:
            return df
        df_null = compute_null_per_col(df)
        columns = df_null[df_null["null_pct"] < self._threshold]["column"].tolist()
        return df[columns]
